"""Define functions related to getting tags."""

import itunespy
import re
from ytmdl.stringutils import (
    remove_multiple_spaces, remove_punct, compute_jaccard, remove_stopwords,
    check_keywords
)
from ytmdl import defaults
from simber import Logger
from ytmdl.meta import (
    gaana, deezer, saavn, lastfm, musicbrainz, spotify,
    preconfig
)
from unidecode import unidecode

logger = Logger('metadata')


def _logger_provider_error(exception, name):
    """Show error if providers throw an error"""
    logger.debug('{}'.format(exception))
    logger.error(
        "Something went wrong with {}. The program will continue with"
        "the other providers. Please check '{}' for more details.\
            ".format(name, logger.get_log_file()))


def get_from_itunes(SONG_NAME):
    """Try to download the metadata using itunespy."""
    # Try to get the song data from itunes
    try:
        # Get the country from the config
        country = defaults.DEFAULT.ITUNES_COUNTRY
        SONG_INFO = itunespy.search_track(SONG_NAME, country=country)
        return SONG_INFO
    except Exception as e:
        _logger_provider_error(e, 'iTunes')
        return None


def get_from_gaana(SONG_NAME):
    """Get some tags from gaana."""
    try:
        nana = gaana.searchSong(SONG_NAME)
        return nana
    except Exception as e:
        _logger_provider_error(e, 'Gaana')
        return None


def get_from_deezer(SONG_NAME):
    """Get some tags from deezer."""
    try:
        songs = deezer.searchSong(SONG_NAME)
        return songs
    except Exception as e:
        _logger_provider_error(e, 'Deezer')


def get_from_lastfm(SONG_NAME):
    """Get metadata from Last FM"""
    try:
        songs = lastfm.searchSong(SONG_NAME)
        return songs
    except Exception as e:
        _logger_provider_error(e, 'LastFM')


def get_from_saavn(SONG_NAME):
    """
    Get the songs from JioSaavn
    """
    try:
        results = saavn.search_query(SONG_NAME)
        return results
    except Exception as e:
        _logger_provider_error(e, "Saavn")
        return None


def get_from_musicbrainz(SONG_NAME):
    """Get the songs from musicbrainz"""
    try:
        results = musicbrainz.search_song(SONG_NAME)
        return results
    except Exception as e:
        _logger_provider_error(e, "MusicBrainz")
        return None


def get_from_spotify(SONG_NAME):
    """
    Get the songs from Spotify
    """
    try:
        country = defaults.DEFAULT.SPOTIFY_COUNTRY
        logger.debug(f"Using {country} for Spotify country")

        return spotify.search_song(SONG_NAME, country=country)
    except Exception as e:
        _logger_provider_error(e, "Spotify")
        return None


def lookup_from_itunes(ID):
    """Lookup metadata by id using itunespy."""
    # Try to get the song data from itunes
    try:
        # Get the country from the config
        country = defaults.DEFAULT.ITUNES_COUNTRY
        SONG_INFO = itunespy.lookup_track(int(ID), country=country)

        # Only keep track results
        SONG_INFO = [i for i in SONG_INFO if i.type == 'track']
        return SONG_INFO
    except Exception as e:
        _logger_provider_error(e, 'iTunes')
        return None


def lookup_from_spotify(id):
    """
    Lookup the track using the ID on Spotify.
    """
    try:
        country = defaults.DEFAULT.SPOTIFY_COUNTRY
        SONG_INFO = spotify.get_track_from_spotify(id=id, country=country)

        return [SONG_INFO]
    except Exception as e:
        _logger_provider_error(e, 'Spotify')
        return None


def _search_tokens(song_name, song_list):
    """Search song in the cache based on simple each word matching."""
    song_name = remove_punct(
        remove_stopwords(
            remove_multiple_spaces(unidecode(song_name)).lower()
        ))
    tokens1 = song_name.split()
    cached_songs = song_list

    res = []
    for song in cached_songs:
        song_back = song
        name = song.track_name.lower()
        # If there is a part like (featuring ..) or any extra data
        # we should remove it as it doesn't aid the search
        name = re.sub(r'\([^)]*\)', '', name)
        name = remove_stopwords(name)
        name = remove_punct(name)
        name = remove_multiple_spaces(name)
        name = unidecode(name)
        tokens2 = name.split()
        match = check_keywords(tokens1, tokens2)
        if match:
            dist = compute_jaccard(tokens1, tokens2)
            if dist >= preconfig.CONFIG().SEARCH_SENSITIVITY:
                res.append((song_back, dist))
    res = sorted(res, key=lambda x: x[1], reverse=True)

    # Return w/o the dist values
    for i in range(0, len(res)):
        res[i] = res[i][0]
    return res


def filterSongs(data, filters=[]):
    """Filter the songs according to the passed filters.

    In the passed filters the first element is artist.
    The second element is album."""

    # In some cases the data can be None, then just return
    if data is None:
        return data

    new_tuple = []
    rest = []

    for songData in data:
        artistMatch = True
        albumMatch = True

        if filters[0] is not None:
            artistMatch = (songData.artist_name == filters[0])
        if filters[1] is not None:
            albumMatch = (songData.collection_name == filters[1])

        if artistMatch and albumMatch:
            new_tuple.append(songData)
        else:
            rest.append(songData)
    return (new_tuple + rest)


def _extend_to_be_sorted_and_rest(provider_data, to_be_sorted, rest, filters):
    """Create the to be sorted and rest lists"""
    # Before passing for sorting filter the songs
    # with the passed args
    if filters:
        provider_data = filterSongs(provider_data, filters)
    if provider_data is not None:
        to_be_sorted.extend(provider_data[:10])
        rest.extend(provider_data[10:])


def SEARCH_SONG(q="Tera Buzz", filters=[], disable_sort=False):
    """Do the task by calling other functions."""
    to_be_sorted = []
    rest = []

    metadata_providers = defaults.DEFAULT.METADATA_PROVIDERS

    GET_METADATA_ACTIONS = {
        'itunes': get_from_itunes,
        'gaana': get_from_gaana,
        'deezer': get_from_deezer,
        'saavn': get_from_saavn,
        'lastfm': get_from_lastfm,
        'musicbrainz': get_from_musicbrainz,
        'spotify': get_from_spotify
    }

    broken_provider_counter = 0

    for provider in metadata_providers:
        if provider in GET_METADATA_ACTIONS:
            logger.debug(f"Searching metadata with {provider}")
            data_provider = GET_METADATA_ACTIONS.get(
                provider, lambda _: None)(q)
            if data_provider:
                _extend_to_be_sorted_and_rest(
                    data_provider, to_be_sorted, rest, filters)
        else:
            logger.warning(
                '"{}" isn\'t implemented. Skipping!'.format(provider)
            )
            broken_provider_counter += 1

    # to_be_sorted will be empty and it will return None anyway, no need
    # to do it here as well
    if broken_provider_counter == len(metadata_providers):
        logger.critical("{}".format(
            'No metadata provider in the configuration is '
            'implemented. Please change it to something \
                            available or use the --skip-meta flag'))

    if not to_be_sorted:
        return None

    # If sorting is disabled, return as is
    if disable_sort:
        return to_be_sorted

    # Send the data to get sorted
    sorted_data = _search_tokens(q, to_be_sorted)

    # Add the unsorted data
    sorted_data += rest

    return sorted_data


if __name__ == '__main__':
    n = SEARCH_SONG("Cradles", ["Sub Urban", None])

    for i in n:
        print(i.track_name + ' by ' + i.artist_name + ' of ' + i.collection_name)



"""
This module is taken from the repo `https://github.com/cyberboysumanjay/JioSaavnAPI`

author: cyberboysumanjay (https://github.com/cyberboysumanjay)
"""

import urllib3
from traceback import print_exc
import json
from pyDes import *
import base64
from bs4 import BeautifulSoup
import requests


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

des_cipher = des(b"38346591", ECB, b"\0\0\0\0\0\0\0\0",
                 pad=None, padmode=PAD_PKCS5)
base_url = 'https://h.saavncdn.com'
json_decoder = json.JSONDecoder()
headers = {
    'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0) Gecko/20100101 Firefox/49.0'
}


def search_from_song_id(song_id):
    song_base_url = "https://www.jiosaavn.com/api.php?cc=in&_marker=0%3F_marker%3D0&_format=json&model=Redmi_5A&__call=song.getDetails&pids=" + \
        str(song_id)
    song_response = requests.get(song_base_url)
    songs_json = list(filter(lambda x: x.startswith(
        "{"), song_response.text.splitlines()))[0]
    songs_json = json.loads(songs_json)
    try:
        songs_json[song_id]['media_url'] = generate_media_url(
            songs_json[song_id]['media_preview_url'])
    except KeyError:
        songs_json[song_id]['media_url'] = decrypt_url(
            songs_json[song_id]['encrypted_media_url'])
    songs_json[song_id]['image'] = fix_image_url(songs_json[song_id]['image'])
    songs_json[song_id]['song'] = fix_title(songs_json[song_id]['song'])
    songs_json[song_id]['album'] = fix_title(songs_json[song_id]['album'])
    return songs_json[song_id]


def search_from_query(query):
    base_url = f"https://www.saavn.com/api.php?__call=autocomplete.get&_marker=0&query={query}&ctx=android&_format=json&_marker=0"
    response = requests.get(base_url)
    songs_json = list(filter(lambda x: x.startswith(
        "{"), response.text.splitlines()))[0]
    songs_json = json.loads(songs_json)
    songs_data = songs_json['songs']['data']
    songs = []
    for song in songs_data:
        song_id = song['id']
        song_base_url = "https://www.jiosaavn.com/api.php?cc=in&_marker=0%3F_marker%3D0&_format=json&model=Redmi_5A&__call=song.getDetails&pids="+song_id
        song_response = requests.get(song_base_url)
        songs_json = list(filter(lambda x: x.startswith(
            "{"), song_response.text.splitlines()))[0]
        songs_json = json.loads(songs_json)
        try:
            songs_json[song_id]['media_url'] = generate_media_url(
                songs_json[song_id]['media_preview_url'])
        except KeyError:
            songs_json[song_id]['media_url'] = decrypt_url(
                songs_json[song_id]['encrypted_media_url'])
        songs_json[song_id]['image'] = fix_image_url(
            songs_json[song_id]['image'])
        songs_json[song_id]['song'] = fix_title(songs_json[song_id]['song'])

        songs_json[song_id]['album'] = fix_title(songs_json[song_id]['album'])
        songs_json[song_id]['media_url'] = check_media_url(
            songs_json[song_id]['media_url'])
        songs.append(songs_json[song_id])
    return songs


class SaavnSong():
    """Class to store saavn song tags."""

    def __init__(self, SONG):
        """SONG is supposed to be a dict."""
        self.track_name = SONG['song']
        self.release_date = self._get_proper_date(SONG['release_date'])
        self.artist_name = SONG['primary_artists']
        self.collection_name = SONG['album']
        self.primary_genre_name = SONG['language']
        self.track_number = '1'
        self.artwork_url_100 = self._get_proper_img_url(SONG['image'])
        self.track_time = self._convert_time(SONG['duration'])

    def _convert_time(self, duration):
        in_min = int(duration)
        in_time = int(in_min / 60) + (0.01 * (in_min % 60))
        return in_time

    def _get_proper_date(self, date):
        """
        The passed date will be of the form
        YYYY-MM-DD

        We need to convert it to a timezone aware
        form.
        Preferably YYYY-MM-DDTHH:MM:SSZ
        """
        return date + 'T00:00:00Z'

    def _get_proper_img_url(self, url):
        """
        The URL provided by Saavn will be a 500x500

        We need to return it in 100x100
        """
        return url.replace("500x500", "100x100")


def search_query(query):
    """
    Just a wrapper over the search_from_query method

    This method will return the data in the format Ytmdl wants it
    in. This will have the same attributes that all the other results
    of ytmdl meta providers have.
    """
    results = [SaavnSong(song) for song in search_from_query(query)]
    return results


def generate_media_url(url):
    url = url.replace("preview", "h")
    url = url.replace("_96_p.mp4", "_320.mp3")
    return url


def get_song_id(query):
    url = query
    songs = []
    try:
        res = requests.get(url, headers=headers, data=[('bitrate', '320')])
        id = res.text.split('"song":{"type":"')[1].split(
            '","image":')[0].split('"id":"')[-1]
        return id
    except Exception as e:
        print_exc()
    return None


def getAlbum(albumId):
    songs_json = []
    try:
        response = requests.get(
            'https://www.jiosaavn.com/api.php?_format=json&__call=content.getAlbumDetails&albumid={0}'.format(albumId), verify=False)
        if response.status_code == 200:
            songs_json = list(filter(lambda x: x.startswith(
                "{"), response.text.splitlines()))[0]
            songs_json = json.loads(songs_json)
            songs_json['name'] = fix_title(songs_json['name'])
            songs_json['image'] = fix_image_url(songs_json['image'])
            for songs in songs_json['songs']:
                try:
                    songs['media_url'] = generate_media_url(
                        songs['media_preview_url'])
                except KeyError:
                    songs['media_url'] = decrypt_url(
                        songs['encrypted_media_url'])
                songs['image'] = fix_image_url(songs['image'])
                songs['song'] = fix_title(songs['song'])
                songs['album'] = fix_title(songs['album'])
            return songs_json
    except Exception as e:
        print_exc()
        return None


def AlbumId(input_url):
    try:
        headers = setProxy()
        res = requests.get(input_url, headers=headers)
        if 'internal error' in res.text:
            return None
        else:
            id = res.text.split('"album_id":"')[1].split('"')[0]
            return id
    except Exception:
        print_exc()
        return None


def setProxy():
    base_url = 'https://h.saavncdn.com'
    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0) Gecko/20100101 Firefox/49.0'
    }
    return headers


def getPlayList(listId):
    try:
        response = requests.get(
            'https://www.jiosaavn.com/api.php?listid={0}&_format=json&__call=playlist.getDetails'.format(listId), verify=False)
        if response.status_code == 200:
            response_text = (response.text.splitlines())
            songs_json = list(
                filter(lambda x: x.endswith("}"), response_text))[0]
            songs_json = json.loads(songs_json)
            songs_json['firstname'] = fix_title(songs_json['firstname'])
            songs_json['listname'] = fix_title(songs_json['listname'])
            songs_json['image'] = fix_image_url(songs_json['image'])
            for songs in songs_json['songs']:
                songs['image'] = fix_image_url(songs['image'])
                songs['song'] = fix_title(songs['song'])
                songs['album'] = fix_title(songs['album'])
                try:
                    songs['media_url'] = generate_media_url(
                        songs['media_preview_url'])
                except KeyError:
                    songs['media_url'] = decrypt_url(
                        songs['encrypted_media_url'])
            return songs_json
        return None
    except Exception:
        print_exc()
        return None


def getListId(input_url):
    headers = setProxy()
    res = requests.get(input_url, headers=headers).text
    return res.split('"type":"playlist","id":"')[1].split('"')[0]


def getSongsJSON(listId):
    url = 'https://www.jiosaavn.com/api.php?listid=' + \
        str(listId)+'&_format=json&__call=playlist.getDetails'
    response_json = requests.get(url).text
    struct = {}
    try:  # try parsing to dict
        dataform = str(response_json).split('-->')[-1]
        struct = json.loads(dataform)
    except Exception:
        print_exc()
        return None
    return struct


def decrypt_url(url):
    enc_url = base64.b64decode(url.strip())
    dec_url = des_cipher.decrypt(enc_url, padmode=PAD_PKCS5).decode('utf-8')
    dec_url = dec_url.replace("_96.mp4", "_320.mp3")
    return dec_url


def fix_title(title):
    title = title.replace('&quot;', '')
    return title


def fix_image_url(url):
    url = str(url)
    if 'http://' in url:
        url = url.replace("http://", "https://")
    url = url.replace('150x150', '500x500')
    return url


def expand_url(url):
    try:
        session = requests.Session()
        resp = session.head(url, allow_redirects=True)
        return(resp.url)
    except Exception as e:
        print("URL Redirect Error: ", e)
        return url


def check_media_url(dec_url):
    ex_dec_url = expand_url(dec_url)
    r = requests.head(ex_dec_url)
    if r.status_code != 200:
        fixed_dec_url = dec_url.replace(".mp3", '.mp4')
        fixed_dec_url = expand_url(fixed_dec_url)
        return fixed_dec_url
    return ex_dec_url



"""Contain all the core functions for ytmdl"""

from typing import Union
from simber import Logger
from colorama import Fore, Style

from ytmdl import (
    yt,
    song,
    stringutils,
    defaults,
    utility,
    manual,
    metadata
)
from ytmdl.exceptions import (
    DownloadError, ConvertError, NoMetaError, MetadataError
)


logger = Logger("core")


def search(song_name, args) -> Union[str, str]:
    """Search the song on YouTube, ask the user for an input and accordingly
    return a selected song.

    The song can be extracted either from an URL or the name. If the name is
    present, we will search on YouTube for the name, ask the user for a choice
    and accordingly return the choice.

    If the URL is present, we will extract the data from the URL and return it
    accordingly.
    """
    # Declare some default values
    YOUTUBE_LINK_BASE = "https://youtube.com{}"
    PASSED_CHOICE = args.choice
    URL = args.url
    IS_QUIET = args.quiet

    if URL is None:
        if IS_QUIET:
            logger.info('Quiet is enabled')

        logger.info('Searching Youtube for {}{}{}'.format(
            Fore.LIGHTYELLOW_EX,
            song_name,
            Style.RESET_ALL
        ))

        data = yt.search(song_name, not args.disable_metaadd,
                         args.proxy,
                         kw=[args.artist, args.album])

        # Handle the exception if urls has len 0
        if len(data) == 0:
            logger.critical(
                "No song found. Please try again with a different keyword.")

        if not IS_QUIET:
            # Ask for a choice
            choice = song.getChoice(data, 'mp3')
        elif PASSED_CHOICE is not None and PASSED_CHOICE <= len(data):
            choice = PASSED_CHOICE - 1
            logger.info("Using {} as choice".format(PASSED_CHOICE))
        else:
            # Extract the verified music result if valid
            choice = song.get_default(data) - 1

        # Check if choice if -2. If it is that, then we need to stop executing
        # of the current song and gracefully exit.
        if choice == -2:
            return False, False

        return (
            YOUTUBE_LINK_BASE.format(data[choice]["href"]),
            data[choice]["title"]
        )

    # If the url is passed then get the data
    data = []

    # Strip unwanted stuff from the URL
    URL = stringutils.srtip_unwanted_words_from_url(URL)

    # Get video data from youtube
    temp_data = yt.scan_video(yt.get_href(URL), args.proxy)

    # Sometimes the temp_data may be returned as unauthorized, skip that
    if type(temp_data) is str and temp_data.lower() == "Unauthorized".lower():
        if args.ignore_errors:
            logger.warning("{}: is unauthorized".format(URL))
            return
        else:
            logger.critical("{}: is unauthorized".format(URL))

    data.append(temp_data)

    # In this case choice will be 0
    return URL, data[0]["title"]


def download(link, yt_title, args) -> str:
    """Download the song by using the passed link.

    The song will be saved with the passed title.
    Return the saved path of the song.
    """
    logger.info('Downloading {}{}{} in {}{}kbps{}'.format(
        Fore.LIGHTMAGENTA_EX,
        yt_title,
        Style.RESET_ALL,
        Fore.LIGHTYELLOW_EX,
        defaults.DEFAULT.SONG_QUALITY,
        Style.RESET_ALL
    ))
    path = yt.dw(link, args.proxy, yt_title, args.format)

    if type(path) is not str:
        # Probably an error occured
        raise DownloadError(link, path)

    logger.info('Downloaded!')
    return path


def convert(
    path: str,
    passed_format: str,
    start: float = None,
    end: float = None
) -> str:
    """Convert the song into the proper format as asked by
    the user.
    """
    FORMAT_CONVERSION_MAP = {
        "mp3": utility.convert_to_mp3,
        "opus": utility.convert_to_opus
    }

    # We need to check if start and end are passed.
    # If those are passed it means only a part of the song is
    # to be extracted.
    logger.debug("{}:{}".format(start, end))
    if start is not None and end is not None:
        conv_name = utility.extract_part_convert(
            path, passed_format, start, end)

        # We need to raise exception if something went wrong
        if type(conv_name) is not str:
            raise ConvertError(conv_name)

        return conv_name

    # If it is m4a, don't convert
    if passed_format == "m4a":
        return path

    # Else the format needs to be in the list
    # It should probably in the list since the check
    # is done once before by the main function
    if passed_format not in FORMAT_CONVERSION_MAP.keys():
        return

    conv_name = FORMAT_CONVERSION_MAP.get(passed_format)(path)

    if type(conv_name) is not str:
        raise ConvertError(conv_name)

    return conv_name


def trim(name: str, args) -> None:
    """Trim the song of unwanted noise by making calls to the
    internal functions.
    """
    # Check if we need to import trim. Importing it is realy inefficient if
    # the user is not going to use it.
    if not args.trim:
        return

    logger.debug("Need to trim the song, importing the trim module.")
    try:
        from ytmdl import trim
    except ImportError:
        logger.error("Dependencies needed for trim are not installed. "
                     "Please use the [noise-clean] specifier when "
                     "installing ytmdl. The script will continue "
                     "without trimming.")
        return

    # Trim the song if the trim option is passed.
    logger.info("Passing the song to get trimmed.")
    trim.Trim(name)


def meta(conv_name: str, song_name: str, search_by: str, args):
    """Handle adding the metadata for the passed song.

    We will use the passed name to search for metadata, ask
    the user for a choice and accordingly add the meta to
    the song.
    """
    PASSED_FORMAT = args.format
    IS_QUIET = args.quiet

    # Check if the user wants a new name for metadata
    if args.ask_meta_name:
        search_by = utility.get_new_meta_search_by(search_by)

    if args.manual_meta:
        # Read the values from the user.
        TRACK_INFO = manual.get_data(song_name)

        # Since above code will return a list with just
        # one element, the option will be set to 0 by
        # default and won't ask the user
    elif args.itunes_id:
        logger.info('Direct iTunes lookup for {}...'.format(args.itunes_id))
        TRACK_INFO = metadata.lookup_from_itunes(args.itunes_id)
    elif args.spotify_id:
        logger.info('Direct Spotify lookup for {}...'.format(args.spotify_id))
        TRACK_INFO = metadata.lookup_from_spotify(args.spotify_id)
    else:
        # Else add metadata in ordinary way
        logger.info('Getting song data for {}...'.format(search_by))
        TRACK_INFO = metadata.SEARCH_SONG(search_by, filters=[
                                          args.artist, args.album],
                                          disable_sort=args.disable_sort)

    # If no meta was found raise error
    if not TRACK_INFO:
        # Check if we are supposed to add manual meta
        if args.on_meta_error != "manual":
            raise NoMetaError(search_by)

        TRACK_INFO = manual.get_data(song_name)
        return TRACK_INFO

    logger.info('Setting data...')
    option = song.setData(TRACK_INFO, IS_QUIET, conv_name, PASSED_FORMAT,
                          args.choice)

    if type(option) is not int:
        raise MetadataError(search_by)

    return TRACK_INFO[option]



"""All directory handling definitions."""

import os
import glob
import shutil
from html import unescape
from re import sub

from ytmdl import defaults
from simber import Logger

logger = Logger("Dir")


def __replace_special_characters(passed_name: str) -> str:
    """
    In the passed name, replace the special characters like
    / with a `-` so that it does not raise any errors
    related to the OS while moving the file
    """
    # TODO: Also add support for removing backslash
    return sub(r'/', '-', passed_name)


def cleanup(TRACK_INFO, index, datatype, remove_cached=True):
    """Move the song from temp to $HOME/Music dir."""
    try:
        SONG = glob.glob(os.path.join(
            defaults.DEFAULT.SONG_TEMP_DIR,
            '*{}'.format(datatype)
        ))
        SONG = SONG[0]

        SONG_NAME = os.path.basename(SONG)

        DIR = defaults.DEFAULT.SONG_DIR
        logger.debug(DIR)

        # Check if DIR has $ in its path
        # If it does then make those folders accordingly

        if '$' in DIR:
            DIR, name = make_custom_dir(DIR, TRACK_INFO[index])

            if name is not None:
                os.rename(SONG, name + '.mp3')
                SONG_NAME = name + '.mp3'
                SONG = SONG_NAME

        dest_filename = os.path.join(
            DIR, __replace_special_characters(SONG_NAME))
        shutil.move(SONG, dest_filename)

        if remove_cached:
            _delete_cached_songs(datatype)

        logger.info('Moved to {}...'.format(DIR))
        return True
    except Exception as e:
        logger.critical("Failed while moving with error: {}".format(e))
        return False


def _delete_cached_songs(ext='mp3'):
    """Delete cached songs"""
    # We need to call this after song is moved
    # because otherwise if there is an error along the way
    # next time a wrong song may be copied.
    SONGS_PATH = os.path.join(
        defaults.DEFAULT.SONG_TEMP_DIR,
        '*{}'.format(ext)
    )
    deleted = False
    for song in glob.glob(SONGS_PATH):
        deleted = True
        os.remove(song)
        logger.debug('Removed "{}" from cache'.format(os.path.basename(song)))
    if deleted:
        logger.debug('{}'.format('Deleted cached songs'))


def ret_proper_names(ordered_names):
    """Return a list with the names changed to itunespy supported ones.

    For eg: Artist to artist_name
    """
    info_dict = {'Artist': 'artist_name',
                 'Title': 'track_name',
                 'Album': 'collection_name',
                 'Genre': 'primary_genre_name',
                 'TrackNumber': 'track_number',
                 'ReleaseDate': 'release_date'
                 }

    logger.debug(ordered_names)
    logger.debug(info_dict)

    new_names = []
    for name in ordered_names:
        new_names.append(info_dict.get(name))

    return new_names


def seperate_kw(uns_kw):
    """Separate the keywords and return a list."""
    sep_kw = []

    # Check if -> is present in the name
    if '->' not in uns_kw:
        sep_kw.append(uns_kw)
    else:
        while '->' in uns_kw:
            pos = uns_kw.find("->")
            sep_kw.append(uns_kw[:pos])
            uns_kw = uns_kw[pos + 2:]

        sep_kw.append(uns_kw)
    return sep_kw


def make_custom_dir(DIR, TRACK_INFO):
    """If the dirname has $ in it then we need to make them.

    The DIR is probably in the format of
    keyword->keyword->keyword
    """
    pos = DIR.index('$')

    # base_DIR is where the folders will be made
    base_DIR = DIR[:pos]

    remaining = DIR[pos + 1:]

    order_dir = seperate_kw(remaining)

    # The last element is to be returned and not considered as
    # a folder
    last_element = order_dir[-1]

    # Replace [] from it
    if last_element[0] == '[' and last_element[-1] == ']':

        last_element = last_element.replace('[', '')
        last_element = last_element.replace(']', '')

        order_dir[-1] = last_element

        order_dir = ret_proper_names(order_dir)

        last_element = order_dir[-1]

    else:
        last_element = None
        order_dir = ret_proper_names(order_dir)

    logger.debug(TRACK_INFO)

    if last_element is not None:
        last_element = getattr(TRACK_INFO, last_element)
        order_dir = order_dir[:len(order_dir) - 1]

    for kw_name in order_dir:
        dir_name = unescape(getattr(TRACK_INFO, kw_name))

        # Sometimes, certain strings have / in the name which creates
        # issues since those strings are used to create directories.
        # Whenever there is a /, replace it with -
        # Sometimes strings also contains [\,?,",<>] which may cause error
        dir_name = sub('[\\\\?<>/"]', "-", dir_name)

        new_dir = os.path.join(base_DIR, dir_name)

        # Make the dir only if it doesn't already exist
        if not os.path.isdir(new_dir):
            os.mkdir(new_dir)

        # Now make the new_dir base_DIR
        base_DIR = new_dir

    return (base_DIR, last_element)


def dry_cleanup(current_path, passed_name):
    """
    Move the song from the current path to the
    song dir and change the name to the passed_name.

    This is only for when the meta-skip option is passed,
    in which case the song needs to be moved from the cache
    to the user directory.
    """
    try:
        extension = os.path.basename(current_path).split(".")[-1]
        logger.debug("ext: {}".format(extension))

        new_basename = "{}.{}".format(passed_name, extension)
        DEST = defaults.DEFAULT.SONG_DIR

        # NOTE: If the DEST is a dynamic directory, then we cannot
        # do a dry cleanup. So we'll have to use the base directory
        # instead.
        if "$" in DEST:
            logger.debug(DEST)

            # pylama:ignore=E501
            logger.warning(
                "Destination is a dynamic directory but this is a dry cleanup. Don't pass `--skip-meta` if you don't want this!")

            # Use the base directory
            DEST = DEST[:DEST.find("$")]

            logger.debug(f"Using {DEST} as destination instead")

        logger.debug("Moving to: {}".format(DEST))

        # Create the destination file name
        dest_filename = os.path.join(
            DEST, __replace_special_characters(new_basename))

        shutil.move(current_path, dest_filename)

        logger.info('Moved to {}...'.format(DEST))
        return True
    except Exception as e:
        logger.error("{}".format(e))
        return False



"""
Handle all the preconfigurations related to metadata
"""

from ytmdl.meta import deezer
from ytmdl.meta import lastfm
from ytmdl.meta import musicbrainz


class CONFIG:
    """
    Class to handle all the preconfigurations
    related to the metadata sources
    """

    # Some sources require fetching extra data
    GET_EXTRA_DATA = {
        'deezer': deezer.get_more_data,
        'lastfm': lastfm.get_more_data,
        'musicbrainz': musicbrainz.get_more_data
    }

    # Search sensitivity is the restriction on matching the
    # search results to the value entered by the user.
    SEARCH_SENSITIVITY = 0.5



"""Handle inserting metadata manually."""

from datetime import datetime
from simber import Logger
from re import sub, match


logger = Logger("manual")


class Meta:
    """
    Meta Class will have properties same as
    those of Gaana and Itunes in order to make
    them compatible with the other modules
    that are using this data to write to the
    song files.

    Following properties will only be present

    release_date        : Date of Release of the song
    track_name          : Name of the song
    artist_name         : Name of the artist(s)
    collection_name     : Name of the album
    primary_genre_name  : Genre of the song
    track_number        : Number of the track in the album
    artwork_url_100     : URL of the album cover
    """
    def __init__(self):
        self.release_date = "{}T00:00:00Z".format(datetime.now().date())
        self.track_name = "N/A"
        self.artist_name = "N/A"
        self.collection_name = "N/A"
        self.primary_genre_name = "N/A"
        self.track_number = "1"
        self.artwork_url_100 = ""

    def _read_individual(self, default_value):
        """
        Read each value and do the usual checks.

        One value is confirmed legit, return it.
        """
        temp_value = input("")
        # Remove starting and terminating spaces
        # and more than one space
        temp_value = sub(r'^\ |\ $', '', temp_value)
        temp_value = sub(r'[\ ]{2,}', ' ', temp_value)

        return temp_value if temp_value != "" else default_value

    def read_values(self):
        """
        Read the values from the user and update them.
        """
        logger.info("Entering interactive mode. Press enter to skip and keep default values.")
        print("Enter date of release of the song (default: {})".format(
                    self.release_date
            ), end=": ")
        self.release_date = self._read_individual(self.release_date)

        print("Enter name of the song (default: {})".format(
                    self.track_name
            ), end=": ")
        self.track_name = self._read_individual(self.track_name)

        print("Enter name of the artist(s) (default: {})".format(
                    self.artist_name
            ), end=": ")
        self.artist_name = self._read_individual(self.artist_name)
        print("Enter name of the album (default: {})".format(
                    self.collection_name
            ), end=": ")
        self.collection_name = self._read_individual(self.collection_name)

        print("Enter genre of the song (default: {})".format(
                    self.primary_genre_name
            ), end=": ")
        self.primary_genre_name = self._read_individual(self.primary_genre_name)

        print("Enter track number (default: {})".format(
                    self.track_number
            ), end=": ")
        self.track_number = self._read_individual(self.track_number)

        print("Enter URL for album cover (default: {})".format(
                    self.artwork_url_100
            ), end=": ")
        self.artwork_url_100 = self._read_individual(self.artwork_url_100)

        logger.info("Leaving interactive mode.")


def get_data(query_name):
    """
    Get the data from the user.
    """
    meta = Meta()
    meta.track_name = query_name
    meta.read_values()
    return [meta]


if __name__ == "__main__":
    meta = Meta()
    meta.read_values()

    print(meta.release_date)
    print(meta.track_name)
    print(meta.collection_name)
    print(meta.artist_name)
    print(meta.primary_genre_name)
    print(meta.track_number)
    print(meta.artwork_url_100)



"""Contains the definition of class DEFAULT."""
from ytmdl import setupConfig
import os
from xdg.BaseDirectory import xdg_cache_home


def _providers_string_to_list(val):
    """Convert string to list if not already"""
    # Use a set to remove duplicates
    if type(val) == str:
        return list(set(val.replace(' ', '').split(',')))
    return list(set(val))


class DEFAULT:
    """DEFAULT class contains value of different constants."""

    def __parse_dir_path(dir_path: str):
        """
        Parse the passed directory path and return a proper
        string. This is for situations when the path might
        have something like a `~` in it.
        """
        if dir_path.startswith("~"):
            dir_path = os.path.expanduser(dir_path)

        return dir_path

    # The home dir
    HOME_DIR = os.path.expanduser('~')

    # The directory where songs will be saved
    SONG_DIR = __parse_dir_path(setupConfig.GIVE_DEFAULT(1, 'SONG_DIR'))

    # The temp directory where songs will be modded
    SONG_TEMP_DIR = os.path.join(xdg_cache_home, 'ytmdl')

    # The path to keep cover image
    COVER_IMG = os.path.join(SONG_TEMP_DIR, 'cover.jpg')

    # The song quality
    SONG_QUALITY = setupConfig.GIVE_DEFAULT(1, 'QUALITY')

    DEFAULT_FORMAT = setupConfig.GIVE_DEFAULT(1, 'DEFAULT_FORMAT')

    # The metadata providers
    METADATA_PROVIDERS = _providers_string_to_list(
        setupConfig.GIVE_DEFAULT(1, 'METADATA_PROVIDERS'))

    VALID_FORMATS = setupConfig.DEFAULTS().VALID_FORMATS

    ON_ERROR_OPTIONS = setupConfig.DEFAULTS().ON_ERROR_OPTIONS
    ON_ERROR_DEFAULT = setupConfig.GIVE_DEFAULT(1, 'ON_META_ERROR')

    ITUNES_COUNTRY = setupConfig.GIVE_DEFAULT(1, 'ITUNES_COUNTRY')

    SPOTIFY_COUNTRY = setupConfig.GIVE_DEFAULT(1, 'SPOTIFY_COUNTRY')



"""Some definitions to interact with the command line."""

import subprocess
from os import remove, path, popen
from ytmdl import defaults, prepend
from shutil import which
import ffmpeg
from simber import Logger
from rich.prompt import Confirm, Prompt

logger = Logger("Utility")


def exe(command):
    """Execute the command externally.

    Written by Nishan Pantha.
    """
    command = command.strip()
    c = command.split()
    output, error = subprocess.Popen(c,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE).communicate()
    output = output.decode('utf-8').strip()
    error = error.decode('utf-8').strip()
    return (output, error)


def get_terminal_length():
    """Return the length of the terminal."""
    rows, cols = popen('stty size', 'r').read().split()

    return int(cols)


def convert_to_mp3r(path):
    """Convert the file to mp3 using ffmpeg."""
    try:
        new_name = path + '_new.mp3'

        command = "ffmpeg -loglevel panic -i {} -vn -ar 44100 -ac 2 -ab {}k -f mp3 {}".format(path,
                                                                                             defaults.DEFAULT.SONG_QUALITY,
                                                                                              new_name)
        output, error = exe(command)

        # Delete the temp file now
        remove(path)
        return new_name
    except Exception as e:
        return e


def convert_to_mp3(path, start=None, end=None, cleanup_after_done=True):
    """Covert to mp3 using the python ffmpeg module."""
    new_name = path + '_new.mp3'
    params = {
        "loglevel": "panic",
        "ar": 44100,
        "ac": 2,
        "ab": '{}k'.format(defaults.DEFAULT.SONG_QUALITY),
        "f": "mp3"
    }

    try:
        if start is not None and end is not None:
            params["ss"] = start
            params["to"] = end

        job = ffmpeg.input(path).output(
                                new_name,
                                **params
                            )
        job.run()

        # Delete the temp file now
        if cleanup_after_done:
            remove(path)

        return new_name
    except ffmpeg._run.Error:
        # This error is usually thrown where ffmpeg doesn't have to
        # overwrite a file.
        # The bug is from ffmpeg, I'm just adding this catch to
        # handle that.
        return new_name


def convert_to_opus(path, start=None, end=None, cleanup_after_done=True):
    """Covert to opus using the python ffmpeg module."""
    new_name = path + '_new.opus'
    params = {
        "loglevel": "panic",
        "f": "opus"
    }

    try:
        if start is not None and end is not None:
            params["ss"] = start
            params["to"] = end

        job = ffmpeg.input(path).output(
                            new_name,
                            **params
                        )
        job.run()

        # Delete the temp file now
        if cleanup_after_done:
            remove(path)

        return new_name
    except ffmpeg._run.Error:
        # This error is usually thrown where ffmpeg doesn't have to
        # overwrite a file.
        # The bug is from ffmpeg, I'm just adding this catch to
        # handle that.
        return new_name


def extract_m4a(path, start=None, end=None, cleanup_after_done=True):
    """Extract a m4a file from the given path based on the start
    and end passed.

    This function is to be called internally only for supporting
    songs with chapters as supported by YouTube.
    """
    if not start and not end:
        logger.error("Cannot trim without start and end")

    new_name = path + '_new.m4a'
    try:
        job = ffmpeg.input(path).output(
                    new_name, ss=start, to=end, loglevel="panic")
        job.run()

        if cleanup_after_done:
            remove(path)

        return new_name
    except ffmpeg._run.Error as e:
        logger.warning(str(e))
        return new_name


def extract_part_convert(path, format, start, end):
    """Extract part of the file using the path provided and accordingly
    convert to the given format.
    """
    FORMAT_MAP = {
        "mp3": convert_to_mp3,
        "opus": convert_to_opus,
        "m4a": extract_m4a
    }

    # Format will be checked by the main function so no need
    # to check here.
    converted_name = FORMAT_MAP.get(format)(path, start, end, False)
    return converted_name


def is_valid(dir_path):
    """Check if passed path is valid or not."""
    if not path.isfile(dir_path):
        return False
    else:
        return True


def get_songs(file_path):
    """Extract the songs from the provided list."""

    if is_valid(file_path):
        RSTREAM = open(file_path, 'r')

        song_tup = RSTREAM.read().split("\n")

        return song_tup
    else:
        return []


def is_present(app):
    """Check if the passed app is installed in the machine."""
    return which(app) is not None


def get_new_title(old_title):
    """
    Ask the user if they would like to go with the old
    title passed as an argument.

    If they say so, it will be returned as it is.

    However, if they want to change the title, get the new
    title and return that one.
    """
    logger.info(
            "Current extracted title for the song is: `{}`".format(old_title)
        )
    logger.info(
            "Most extracted titles are not accurate and they affect the meta search"
            )

    is_change = Confirm.ask("Would you like to change", default=True)

    if not is_change:
        return old_title

    # Else ask for new title
    title = Prompt.ask("Enter new title")
    return title


def get_new_meta_search_by(old_search_by):
    """
    Ask the user to enter a new title to search the metadata with.

    This function will be called when the user explicitly asks for a
    different title to search the metadata with.
    """
    prepend.PREPEND(1)
    new_search_by = Prompt.ask("Enter new title to search metadata with",
                               default=old_search_by)

    if new_search_by == old_search_by:
        logger.info("You can disable this behaviour by removing the"
                    "`--ask-meta-name` flag.")

    return new_search_by


if __name__ == "__main__":
    print(get_new_title("Haha"))



"""
Python module to search gaana.com using their API.

Uses api.gaana.com to get search results.
"""

import requests

# Define the base url
base_url = "http://api.gaana.com/?type=search&subtype=search_song&key={}&token=b2e6d7fbc136547a940516e9b77e5990&format=JSON"

SONG = []


class GaanaSongs():
    """Class to store gaana song tags."""

    def __init__(self, SONG):
        """SONG is supposed to be a dict."""
        self.track_name = SONG['track_title']
        self.release_date = SONG['release_date']
        self.artist_name = SONG['artist'][0]['name']
        self.provider = "gaana"
        self.collection_name = SONG['album_title']
        self.primary_genre_name = SONG['gener'][0]['name']
        self.track_number = '1'
        self.artwork_url_100 = SONG['artwork_large']
        self.track_time = self._convert_time(SONG['duration'])

    def _convert_time(self, duration):
        in_min = int(duration)
        in_time = int(in_min / 60) + (0.01 * (in_min % 60))
        return in_time


def searchSong(querry, lim=40):
    """Nanan."""
    url = base_url.format(querry)
    r = requests.get(url)
    data = r.json()
    data = data['tracks']
    SONG_TUPLE = []

    if data:
        for i in range(0, len(data)):
            song_obj = GaanaSongs(data[i])
            SONG_TUPLE.append(song_obj)

    return SONG_TUPLE


if __name__ == '__main__':
    q = input("Enter the querry: ")
    dat = searchSong(q)
    for i in range(0, len(dat)):
        print(dat[i].collection_name)



"""
Handle the archive related feature.
"""

from pathlib import Path
from typing import List, Union
from simber import Logger
from io import TextIOWrapper

from ytmdl.yt import extract_video_id


# Create logger
logger = Logger("archive")


def open_archive_stream(file: str) -> Union[List, TextIOWrapper]:
    """
    Read the archive file and extract all the videoId's
    passed. This file will be read as text and should contain
    videoId's in each line.
    """
    file_path: Path = Path(file).expanduser()

    # Check if the file exists
    if not file_path.exists():
        logger.critical("Passed archive file does not exist. Exiting!")

    stream: TextIOWrapper = file_path.open("r+")
    file_content: List = stream.read().split("\n")
    return file_content, stream


def is_present_in_archive(file_content: List, youtube_link: str) -> bool:
    """
    Check if the passed song is present
    in the download-archive file passed
    and accordingly return.
    """
    video_id: str = extract_video_id(youtube_link)

    return video_id in file_content


def add_song_to_archive(stream: TextIOWrapper, youtube_link: str) -> None:
    """
    Add the passed song to the archive file.

    This method is supposed to be called when the song
    is considered `worked on` and not just when the song
    is just downloaded.
    """
    video_id: str = extract_video_id(youtube_link)

    # Go to the end of the file
    stream.seek(0, 2)
    stream.write(f"\n{video_id}")



"""
Handle the stuff related to trimming the song
in order to remove the speech.
"""

from inaSpeechSegmenter import Segmenter
from warnings import filterwarnings
from os import environ, remove, rename
import ffmpeg
from simber import Logger

# Setup the logger name
logger = Logger("Trimmer")


# Remove the warnings from tensorflow
environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# Ignore warnings
filterwarnings("ignore")


class Trim:
    """
    Class that will handle the processing of the files.

    filename: File that we want to process.

    We will run the speech segmenter on it and get
    the music timestamps. Once we have the stamps, we can just
    trim the music part from the song and keep the name same.
    """
    def __init__(self, filename):
        self.filename = filename
        self.segmenter = Segmenter(
                                vad_engine='smn',
                                detect_gender=False,
                                ffmpeg='ffmpeg'
                            )
        self._find_music()
        self._trim()

    def _find_music(self):
        """
        Call the segmenter on the file and extract the
        music timestamps from there.
        """
        segmentation = self.segmenter(self.filename)
        logger.debug("Segmentation tuple: {}".format(segmentation))

        # Only keep all the music stamps
        segmentation = [stamp for stamp in segmentation if stamp[0] == "music"]

        if not len(segmentation):
            logger.critical(
                "Could not find music in the file. Try disabling trimming!"
            )

        # We want to consider just the last time stamp tuple,
        # Mostly there are just one time stamp with music, but sometimes
        # it can be more than one.
        segmentation = segmentation[-1]
        self.start_time = segmentation[1]
        self.end_time = segmentation[2]

        logger.debug("ST: {}\nET: {}".format(self.start_time, self.end_time))

    def _trim(self):
        """
        Use ffmpeg to trim the song to the time stamps calculated.

        Make the changes to a temp file.
        Once the work is done remove the original file and rename
        the temporary to original one.
        """
        logger.info("Trimming the song to the found time stamps")

        # Create the temp file name
        temp_name = "{}_temp.{}".format(
                                    self.filename,
                                    self.filename.split(".")[-1]
                                )

        ffmpeg.input(self.filename).output(
                                        temp_name,
                                        loglevel="panic",
                                        ss=self.start_time,
                                        to=self.end_time
                    ).run()

        # Once that's done, remove the original file
        remove(self.filename)
        # Now rename the file.
        rename(temp_name, self.filename)
        logger.info("Trimmed the file succesfully!")



"""song.py - Used for song related functions.

All the functions used to interact with the downloaded song are defined here.
"""

from colorama import Fore, Style
from mutagen.id3 import (
    ID3,
    APIC,
    TIT2,
    TPE1,
    TALB,
    TCON,
    TRCK,
    TYER,
    PictureType
)
from mutagen.mp3 import MP3
from mutagen.mp4 import MP4, MP4Cover
from mutagen import File
from mutagen.flac import Picture
from base64 import b64encode
import requests
import os
from rich.prompt import IntPrompt

from ytmdl import prepend, defaults
from simber import Logger
from ytmdl.meta import preconfig
# import traceback

logger = Logger("song")

# ----------------------cover--------------------


def dwCover(song):
    """Download the song cover img from itunes."""
    # Try to download the cover art as cover.jpg in temp
    logger.info("Preparing the album cover")
    try:
        imgURL = song.artwork_url_100

        # Check if the passed imgURL is a local file
        # this is possible if the metadata was entered manually.
        imgURL = os.path.expanduser(imgURL)
        if os.path.exists(imgURL):
            # Probably a file, read it in binary and extract the data
            # then return.
            content = open(imgURL, "rb").read()
            with open(defaults.DEFAULT.COVER_IMG, 'wb') as f:
                f.write(content)
            return True

        # Else might be an URL
        try:
            # Try to get 512 cover art
            imgURL = imgURL.replace('100x100', '2048x2048')
        except Exception:
            pass

        r = requests.get(imgURL)

        with open(defaults.DEFAULT.COVER_IMG, 'wb') as f:
            f.write(r.content)

        return True
    except TimeoutError:
        prepend.PREPEND(2)
        print('Could not get album cover. Are you connected to internet?\a')
        return False
    except Exception as e:
        logger.warning(
            "Error while trying to download image, skipping!: {}".format(e))
        return False
    else:
        return False

# -----------------------tag----------------------


def print_choice(beg, end, SONG_INFO, type):
    """Print the available choices."""
    # Check if end is more than length of SONG_INFO
    if end > len(SONG_INFO):
        end = len(SONG_INFO)

    while beg != end:
        print(Fore.LIGHTMAGENTA_EX, end='')
        print(' [' + str(beg+1) + '] ', end='')
        print(Style.RESET_ALL, end='')
        print(Fore.LIGHTCYAN_EX, end='')
        if type == 'metadata':
            print(SONG_INFO[beg].track_name, end='')
        if type == 'mp3':
            print('[M] ' if SONG_INFO[beg]['verified_music'] else '', end='')
            print(SONG_INFO[beg]['title'], end='')
        print(Style.RESET_ALL, end='')
        print(' by ', end='')
        print(Fore.YELLOW, end='')
        if type == 'metadata':
            print(SONG_INFO[beg].artist_name, end='')
        if type == 'mp3':
            print(SONG_INFO[beg]['author_name'], end='')
            print(Style.RESET_ALL, end='')
            print(' with dur ', end='')
            print(Fore.GREEN, end='')
            print("{}".format(SONG_INFO[beg]['duration']), end='')
        print(Style.RESET_ALL)

        beg += 1

    # Before exiting print another choice to show more
    if end < len(SONG_INFO):
        print(Fore.LIGHTMAGENTA_EX, end='')
        print(' [0]', end='')
        print(Style.RESET_ALL, end='')
        print(Fore.YELLOW, end='')
        print(' More results')
        print(Style.RESET_ALL, end='')


def get_default(songs, type="mp3") -> int:
    """Get the default song that will be selected if the user
    doesn't select a value.
    """
    # If the user is asked to select metadata, then just return
    # 1
    if type != "mp3":
        return 1

    # Else, we need to find the first verified music video present.
    choice = 1

    for index in range(0, len(songs)):
        if songs[index]['verified_music']:
            choice = index + 1
            break

    return choice


def getChoice(SONG_INFO, type):
    """If more than 1 result from getData then ask for a choice."""
    # Print 5 of the search results
    # In case less, print all

    logger.info('Choose One {}'.format(
        '(One with [M] is verified music)'
        if type == 'mp3' else ''))

    results = len(SONG_INFO)

    if results > 5:
        results = 5

    PRINT_WHOLE = True

    default_choice = get_default(SONG_INFO, type)

    beg = 0
    while True:
        # Print the results first
        if PRINT_WHOLE:
            print_choice(beg, results, SONG_INFO, type)
        prepend.PREPEND(1)
        choice = IntPrompt.ask('Enter Choice', default=default_choice)

        logger.debug(choice)
        choice = int(choice)

        # If the choice is 0 then try to print more results
        # The choice is valid if it is in the range and it is greater than 0
        # We also need to break when the user enters -1 which means the exec
        # will skip the current song
        if (choice <= len(SONG_INFO) and choice > 0) or choice == -1:
            break
        elif choice == 0 and results < len(SONG_INFO):
            PRINT_WHOLE = True
            beg = results
            results = beg + 5
        else:
            PRINT_WHOLE = False

    return choice - 1


def set_MP3_data(song, song_path):
    """
    Set the meta data if the passed data is mp3.
    """
    # A variable to see if cover image was added.
    IS_IMG_ADDED = False

    try:
        SONG_PATH = os.path.join(defaults.DEFAULT.SONG_TEMP_DIR,
                                 song_path)

        audio = MP3(SONG_PATH, ID3=ID3)
        data = ID3(SONG_PATH)

        # Download the cover image, if failed, pass
        if dwCover(song):
            imagedata = open(defaults.DEFAULT.COVER_IMG, 'rb').read()
            data.add(APIC(3, 'image/jpeg', 3, 'Front cover', imagedata))
            # REmove the image
            os.remove(defaults.DEFAULT.COVER_IMG)
            IS_IMG_ADDED = True

        # If tags are not present then add them
        try:
            audio.add_tags()
        except Exception:
            pass

        audio.save()

        logger.debug("Passed song release date: ", song.release_date)

        data.add(TYER(encoding=3, text=song.release_date))
        data.add(TIT2(encoding=3, text=song.track_name))
        data.add(TPE1(encoding=3, text=song.artist_name))
        data.add(TALB(encoding=3, text=song.collection_name))
        data.add(TCON(encoding=3, text=song.primary_genre_name))
        data.add(TRCK(encoding=3, text=str(song.track_number)))

        data.save()

        defaults.DEFAULT.SONG_NAME_TO_SAVE = song.track_name + '.mp3'

        # Rename the downloaded file
        os.rename(SONG_PATH, os.path.join(
            defaults.DEFAULT.SONG_TEMP_DIR,
            defaults.DEFAULT.SONG_NAME_TO_SAVE
        ))

        return IS_IMG_ADDED

    except Exception as e:
        logger.debug("{}".format(e))
        return e, False


def set_M4A_data(song, song_path):
    """
    Set the tags in the m4a file passed.
    """
    cover_added = False

    try:
        SONG_PATH = os.path.join(defaults.DEFAULT.SONG_TEMP_DIR,
                                 song_path)
        audio = MP4(SONG_PATH)

        # Download the cover image, if failed, pass
        if dwCover(song):
            imagedata = open(defaults.DEFAULT.COVER_IMG, 'rb').read()
            audio["covr"] = [MP4Cover(
                imagedata,
                imageformat=MP4Cover.FORMAT_JPEG
            )]
            # REmove the image
            os.remove(defaults.DEFAULT.COVER_IMG)
            cover_added = True

        # If tags are not present then add them
        try:
            audio.add_tags()
        except Exception:
            pass

        audio.save()

        # Add the meta data, the key's can be found at
        # https://mutagen.readthedocs.io/en/latest/api/mp4.html#mutagen.mp4.MP4Tags
        audio["\xa9nam"] = song.track_name
        audio["\xa9alb"] = song.collection_name
        audio["\xa9ART"] = song.artist_name
        audio["\xa9day"] = song.release_date
        audio["\xa9gen"] = song.primary_genre_name

        # NOTE: In m4a files, the track number is of the following format
        # track number / track count
        # However, we don't have track count for all songs, so
        # we'll have to find a fallback for that.
        track_count = song.track_count if hasattr(song, 'track_count') else "1"
        logger.debug("Adding track count")
        logger.debug(f"Count: {track_count}")
        audio["trkn"] = [(int(song.track_number), int(track_count))]

        audio.save()

        defaults.DEFAULT.SONG_NAME_TO_SAVE = song.track_name + '.m4a'

        # Rename the downloaded file
        os.rename(SONG_PATH, os.path.join(
            defaults.DEFAULT.SONG_TEMP_DIR,
            defaults.DEFAULT.SONG_NAME_TO_SAVE
        ))

        return cover_added

    except Exception as e:
        return e


def set_OPUS_data(song, song_path):
    """
    Set the data into an OPUS container according to the
    passed data.
    """
    COVER_ADDED = False

    try:
        SONG_PATH = os.path.join(defaults.DEFAULT.SONG_TEMP_DIR,
                                 song_path)
        mutagen_file = File(SONG_PATH)

        # Try adding the tags container
        try:
            mutagen_file.add_tags()
        except Exception:
            # If exception is thrown, the tags already exist
            pass

        # Clear out the tags from the file
        mutagen_file.clear()

        # Try adding the cover
        if dwCover(song):
            imagedata = open(defaults.DEFAULT.COVER_IMG, 'rb').read()
            picture = Picture()
            picture.data = imagedata
            picture.type = PictureType.COVER_FRONT
            picture.mime = "image/jpeg"
            encoded_data = b64encode(picture.write())
            mutagen_file["metadata_block_picture"] = encoded_data.decode(
                "ascii")

            # Remove the image
            os.remove(defaults.DEFAULT.COVER_IMG)
            COVER_ADDED = True

        # Add the tags now
        # Refer to https://www.programcreek.com/python/example/63675/mutagen.File
        # for more information on it
        mutagen_file["Title"] = song.track_name
        mutagen_file["Album"] = song.collection_name
        mutagen_file["Artist"] = song.artist_name
        mutagen_file["Date"] = song.release_date
        mutagen_file["Genre"] = song.primary_genre_name

        mutagen_file.save()

        defaults.DEFAULT.SONG_NAME_TO_SAVE = song.track_name + '.opus'

        # Rename the downloaded file
        os.rename(SONG_PATH, os.path.join(
            defaults.DEFAULT.SONG_TEMP_DIR,
            defaults.DEFAULT.SONG_NAME_TO_SAVE
        ))

        return COVER_ADDED
    except Exception as e:
        return e


def _get_option(SONG_INFO, is_quiet, choice):
    option = 0
    if len(SONG_INFO) > 1:
        if not is_quiet:
            option = getChoice(SONG_INFO, 'metadata')
        elif choice is not None and choice in range(1, len(SONG_INFO)):
            option = choice
    return int(option)


def setData(SONG_INFO, is_quiet, song_path, datatype='mp3', choice=None):
    """Add the metadata to the song."""

    # Some providers need extra daa from other endpoints,
    # this is where we define which need it and where to get
    # it from

    logger.debug(choice)
    option = _get_option(SONG_INFO, is_quiet, choice)
    logger.debug(option)

    song = SONG_INFO[option]

    get_more_data_dict = preconfig.CONFIG().GET_EXTRA_DATA

    # Try to check if the song object has an attribute provider
    # Deezer has it but other objects don't have it.
    # If the provider is present then fetch extra data accordingly

    if hasattr(song, 'provider') and song.provider in get_more_data_dict:
        song = get_more_data_dict.get(song.provider, lambda _: None)(song)

    if datatype == 'mp3':
        img_added = set_MP3_data(
            song,
            song_path,
        )
    elif datatype == 'm4a':
        img_added = set_M4A_data(
            song,
            song_path,
        )
    elif datatype == 'opus':
        img_added = set_OPUS_data(
            song,
            song_path
        )

    # Show the written stuff in a better format
    prepend.PREPEND(1)
    print('================================')
    print('  || YEAR: ' + song.release_date)
    print('  || TITLE: ' + song.track_name)
    print('  || ARTIST: ' + song.artist_name)
    print('  || ALBUM: ' + song.collection_name)
    print('  || GENRE: ' + song.primary_genre_name)
    print('  || TRACK NO: ' + str(song.track_number))

    if img_added:
        print('  || ALBUM COVER ADDED')

    prepend.PREPEND(1)
    print('================================')

    return option




"""Definition of functions that are used to interact with youtube."""

import requests
import os
from urllib.parse import urlparse, parse_qs
import youtube_dl
from youtube_dl.utils import DownloadError
from re import match
from ytmdl import defaults, utility, stringutils
from downloader_cli.download import Download
import traceback
from sys import stdout
from simber import Logger
from ytmdl.exceptions import ExtractError
from ytmdl.utils.ytmusic import get_title_from_ytmusic
from youtubesearchpython import VideosSearch
from typing import List


logger = Logger("yt")


def get_youtube_streams(url):
    """Get both audio & video stream urls for youtube using youtube-dl.

    PS: I don't know how youtube-dl does the magic
    """
    cli = "youtube-dl -g {}".format(url)
    output, error = utility.exe(cli)
    stream_urls = output.split("\n")

    url = stream_urls[1]
    return url


# Function to be called by ytdl progress hook.
def progress_handler(d):
    d_obj = Download('', '')

    if d['status'] == 'downloading':
        length = d_obj._get_terminal_length()
        time_left = d['eta']
        f_size_disp, dw_unit = d_obj._format_size(d['downloaded_bytes'])

        # Total bytes might not be always passed, sometimes
        # total_bytes_estimate is passed
        try:
            total_bytes = d['total_bytes']
        except KeyError:
            total_bytes = d['total_bytes_estimate']

        percent = d['downloaded_bytes'] / total_bytes * 100
        speed, s_unit, time_left, time_unit = d_obj._get_speed_n_time(
            d['downloaded_bytes'],
            0,
            cur_time=d['elapsed'] - 6
        )

        status = r"%-7s" % ("%s %s" % (round(f_size_disp), dw_unit))
        if d['speed'] is not None:
            speed, s_unit = d_obj._format_speed(d['speed'] / 1000)
            status += r"| %-3s " % ("%s %s" % (round(speed), s_unit))

        status += r"|| ETA: %-4s " % (
            "%s %s" %
            (round(time_left), time_unit))

        status = d_obj._get_bar(status, length, percent)
        status += r" %-4s" % ("{}%".format(round(percent)))

        stdout.write('\r')
        stdout.write(status)
        stdout.flush()


def dw_using_yt(link, proxy, song_name, datatype, no_progress=False):
    """
    Download the song using YTDL downloader and use downloader CLI's
    functions to be used to display a progressbar.

    The function will be called by using hooks.
    """
    if datatype == 'mp3' or datatype == 'opus':
        format_ = 'bestaudio/best'
    elif datatype == 'm4a':
        format_ = 'bestaudio[ext=m4a]'

    ydl_opts = {
        'quiet': True,
        'outtmpl': song_name,
        'format': format_,
        'nocheckcertificate': True,
        'source_address': '0.0.0.0'
    }

    if not no_progress:
        ydl_opts['progress_hooks'] = [progress_handler]

    if proxy is not None:
        ydl_opts['proxy'] = proxy

    ydl = youtube_dl.YoutubeDL(ydl_opts)

    try:
        ydl.download([link])
        return 0
    except Exception as e:
        logger.warning("{}".format(e))
        return e


def dw(
        value,
        proxy=None,
        song_name='ytmdl_temp.mp3',
        datatype='mp3',
        no_progress=False
):
    """
    Download the song.

    The song can be downloaded in various types.

    Default type is mp3 as ytmdl was solely designed to download
    MP3 songs, however due to user requests other formats are
    added.
    """
    # If song_name doesn't have mp3 extension, add it
    if datatype == "mp3" and not song_name.endswith(datatype):
        song_name += '.' + datatype
    elif datatype == "m4a" and not song_name.endswith(datatype):
        song_name += '.' + datatype

    try:
        # Replace the spaces with hashes
        song_name = stringutils.remove_unwanted_chars(song_name)

        # The directory where we will download to.
        dw_dir = defaults.DEFAULT.SONG_TEMP_DIR
        logger.info("Saving the files to: {}".format(dw_dir))

        if not os.path.exists(dw_dir):
            os.makedirs(dw_dir)

        # Name of the temp file
        name = os.path.join(dw_dir, song_name)
        logger.debug(name)

        # Start downloading the song
        status = dw_using_yt(value, proxy, name, datatype, no_progress)

        if status == 0:
            return name
        else:
            return status

    except Exception as e:
        # traceback.print_exception(e)
        return e


def get_href(url):
    """Get the watch? part of the url in case of urls."""
    pos_watch = url.index('/watch?v=')

    part = url[pos_watch:]

    return part


def _is_verified(desc: List) -> bool:
    """Determine if the music video is provided
    by the music company to YouTube or not.

    In other words, this will just indicate if the
    song is directly provided by the music producer.

    This can be determined in a few ways.

    - If the desc contains `Auto-Generated by YouTube`
    - If the desc contains `Provided to YouTube by <company>`
    """
    if desc is None:
        return False

    desc = desc[0]["text"]
    return bool(match(r'^provided.to.youtube.by.*|^auto.generated.by.youtube', desc.lower()))


def search(query, bettersearch, proxy, kw=[], lim=20):
    """
    Search the song name and return results

    Search the passed query using the `youtube_search` module
    and extract the results accordingly and return.
    """
    # Add keywords if better search is enabled
    kw = [kw_ for kw_ in kw if kw_ is not None]

    if bettersearch and len(kw):
        query += '+' + '+'.join(kw)

    # Check if proxy is passed.
    proxies = {}
    if proxy is not None:
        proxies['http'] = proxies['https'] = proxy

    # Replace all the spaces with +
    query = query.replace(' ', '+')

    results = VideosSearch(query, limit=lim).result()["result"]

    if not results:
        logger.warning("No search results found!")

    stripped_results = []

    for video in results:
        data = {}
        data['title'] = video['title']
        data['href'] = f"/watch?v={video['id']}"
        data['author_name'] = video['channel']["name"]
        data['duration'] = video['duration']
        data['verified_music'] = _is_verified(video['descriptionSnippet'])

        stripped_results.append(data)

    return stripped_results


def scan_video(url, proxy):
    """Scan the link of the video and return data and."""
    proxies = {}
    if proxy is not None:
        proxies['http'] = proxy
    try:
        search_tmplt = "https://www.youtube.com/oembed?url={}&format=json"
        search_url = search_tmplt.format(url)
        r = requests.get(search_url, proxies=proxies)

        if r.status_code == 200:
            return r.json()
        else:
            return "Unauthorized"

    except Exception:
        return False


def is_playlist(url):
    """
    Check if the passed URL is a youtube playlist
    URL.
    """
    playlist_part = r"https?://(www\.|music\.)?youtube\.com/playlist\?list=.*?$"
    return match(playlist_part, url)


def get_playlist(
    url,
    proxy,
    playlist_start=None,
    playlist_end=None,
    playlist_items=None
):
    """
    Extract the playlist data and return it accordingly.

    The return result will be a dictionary with the following
    entries
    _type: Type of the entity returned.
    url  : URL of the video
    title: Title of the video.
    """
    ydl_opts = {
        'quiet': True,
        'format': 'bestaudio/best',
        'nocheckcertificate': True,
        'dump_single_json': True,
        'extract_flat': True,
    }

    if proxy is not None:
        ydl_opts['proxy'] = proxy
    if playlist_start is not None:
        ydl_opts['playliststart'] = playlist_start
    if playlist_end is not None:
        ydl_opts['playlistend'] = playlist_end
    if playlist_items is not None:
        ydl_opts['playlist_items'] = playlist_items

    # Extract the info now
    songs = youtube_dl.YoutubeDL(ydl_opts).extract_info(url, False)

    # Put another check to see if the passed URL is a playlist
    try:
        if songs["_type"] != "playlist":
            logger.warning("Passed URL is not a playlist URL")
            return None
    except KeyError:
        pass

    # Return the songs now.
    try:
        return songs["entries"], songs["title"]
    except KeyError:
        logger.warning(
            "Something went wrong while extracting the playlist data."
        )
        return None, None


def __get_title_from_yt(url):
    """
    Return the title of the passed URL.
    """
    ydl_opts = {
        "quiet": True,
        'nocheckcertificate': True,
        'source_address': '0.0.0.0'
    }

    logger.debug(url)

    ydl = youtube_dl.YoutubeDL(ydl_opts)

    try:
        data = ydl.extract_info(url, False)
        return stringutils.remove_yt_words(data["title"])
    except DownloadError:
        raise ExtractError(url)
    except KeyError:
        logger.error("Wasn't able to extract the name of the song.")
        return ""


def extract_video_id(url: str) -> str:
    """
    Extract the video ID from the URL.
    """
    try:
        return parse_qs(urlparse(url=url).query)["v"][0]
    except KeyError:
        raise ExtractError(url)


def get_title(url) -> str:
    """
    Try to get the title of the song.

    This is mostly used when URL is passed or playlist
    links are passed since in those cases the title is
    not explicitly passed.
    """
    # Primarily try to get the title by using Youtube
    # Music.

    # If to verify the title from the user
    verify_title = False

    try:
        title = get_title_from_ytmusic(extract_video_id(url=url))
        return title, verify_title
    except ExtractError:
        logger.debug(f"YtMusic wasn't able to find title for {url}")
        pass

    # Try Youtube as a fallback
    verify_title = True
    title = __get_title_from_yt(url)
    return title, verify_title


def get_chapters(url):
    """Get the chapters of the passed URL.
    """
    ydl_opts = {
        "quiet": True,
        'nocheckcertificate': True,
        'source_address': '0.0.0.0'
    }

    info = youtube_dl.YoutubeDL(ydl_opts).extract_info(url, False)

    return info.get("chapters", None)


if __name__ == '__main__':
    print(defaults.DEFAULT.SONG_QUALITY)



"""
Handle extracting metadata from Spotify

We will use Spotify's API to search for tracks and show them
to the user to be used accordingly.
"""

from spotipy import SpotifyClientCredentials
from spotipy import Spotify

from typing import Dict, List

# Yeah I know, this is not a good idea but
# asking the users for their own credentials is jus
# not good enough.
# NOTE: Raise an issue if you have a better idea
# and NO, I don't wanna ask the user to enter these
# creds.
CLIENT_ID = "a166f23a5637429e8bd819df46fa034e"
CLIENT_SECRET = "bbad17df3cde471ab969f081f2ea8bbb"


class SpotifySong(object):
    """
    Class to store data about the songs fetched from
    Spotify.

    We need to clean the data into an object that can
    be utilized if the user choses to add this song as
    metadata.
    """

    def __init__(self, song: Dict) -> None:
        self.track_name = song["name"]
        self.release_date = song["album"]["release_date"]
        self.artist_name = song["artists"][0]["name"]
        self.provider = "spotify"
        self.collection_name = song["album"]["name"]
        self.primary_genre_name = ""  # Seems spotify doesn't provide genre
        self.track_number = song["track_number"]
        self.artwork_url_100 = song["album"]["images"][0]["url"]
        self.track_time = song["duration_ms"]


def search_song(
    query,
    country: str = "US",
    limit: int = 25
) -> List[SpotifySong]:
    """
    Search the song using the API through spotipy
    and accordingly return the results.
    """
    spotify = Spotify(auth_manager=SpotifyClientCredentials(
        client_id=CLIENT_ID, client_secret=CLIENT_SECRET))

    response = spotify.search(
        f"track:{query}", limit=limit, type="track", market=country)
    items = [SpotifySong(item) for item in response["tracks"]["items"]]

    return items


def get_track_from_spotify(id, country: str = "US"):
    """
    Lookup the metadata by using the ID on spotify
    """
    spotify = Spotify(auth_manager=SpotifyClientCredentials(
        client_id=CLIENT_ID, client_secret=CLIENT_SECRET))

    track = spotify.track(id, market=country)
    return SpotifySong(track)



"""Store all the exceptions related to ytmdl that might
arise during the runtime of the app
"""


class DownloadError(Exception):
    """Exception for download error.

    This exception is solely for those cases when
    the download fails for some reason.
    """
    def __init__(self, link, error) -> None:
        super().__init__()

        self.__message = self.__build_message(link, error)

    def __build_message(self, link, error) -> str:
        """Build the error message."""
        return "Download failed for `{}` with error: {}".format(
            link, error
        )

    def __str__(self) -> str:
        return self.__message


class ConvertError(Exception):
    """Exception for conversion erros.

    This exception is raised whenever the conversion goes wrong,
    mostly while using ffmpeg or something similar.
    """
    def __init__(self, error) -> None:
        super().__init__()

        self.__message = self.__build_message(error)

    def __build_message(self, error) -> str:
        """Build the error message"""
        return "Conversion failed with error: {}".format(error)

    def __str__(self) -> str:
        return self.__message


class NoMetaError(Exception):
    """Exception to be raised when no metadata is found."""
    def __init__(self, song) -> None:
        super().__init__()

        self.__message = self.__build_message(song)

    def __build_message(self, song) -> str:
        """Build the error message"""
        return "No metadata found for `{}`".format(song)

    def __str__(self) -> str:
        return self.__message


class MetadataError(Exception):
    """Exception for metadata related errors while setting metadata

    This is only to be raised when something goes wrong while setting
    the metadata for the song.
    """
    def __init__(self, song) -> None:
        super().__init__()

        self.__message = self.__build_message(song)

    def __build_message(self, song) -> str:
        """Build the error message"""
        return "Something went wrong while setting metadata for `{}`".format(
            song
        )

    def __str__(self) -> str:
        return self.__message


class ExtractError(Exception):
    """
    Exception for errors that arise while extracting any detail
    related to the song.

    This is only to be raise when some extraction related error
    occurs.
    """
    def __init__(self, song) -> None:
        super().__init__()

        self.__message = self.__build_message(song)

    def __build_message(self, song) -> str:
        return "Couldn't extract data for: {}".format(song)

    def __str__(self) -> str:
        return self.__message



"""
Python module to search deezer.com using their API.

Uses api.deezer.com to get search results.
"""

import requests

DEEZER_URL = "https://api.deezer.com/"

search_url = "{}search?q={}"

url_extra_track = "{}track/{}"
url_extra_album = "{}album/{}"

SONG = []


class DeezerSongs():
    """Class to store deezer song tags."""

    def __init__(self, SONG):
        """SONG is supposed to be a dict."""
        self.track_name = SONG['title_short']
        self.artist_name = SONG['artist']['name']
        self.provider = 'deezer'
        self.collection_id = SONG['album']['id']
        self.track_id = SONG['id']
        self.track_number = "1"
        self.collection_name = SONG['album']['title']
        self.artwork_url_100 = SONG['album']['cover_medium']
        self.track_time = self._convert_time(SONG['duration'])

    def _convert_time(self, duration):
        in_min = int(duration)
        in_time = int(in_min / 60) + (0.01 * (in_min % 60))
        return in_time


def get_more_data(song):
    url_track = url_extra_track.format(DEEZER_URL, song.track_id)
    url_album = url_extra_album.format(DEEZER_URL, song.collection_id)

    r = requests.get(url_track)
    data_track = r.json()

    r = requests.get(url_album)
    data_album = r.json()

    song.primary_genre_name = data_album['genres']['data'][0]['name']
    song.track_number = data_track['track_position']
    song.release_date = data_track['release_date']

    return song


def searchSong(query, lim=40):
    """Deezer"""
    url = search_url.format(DEEZER_URL, query)
    r = requests.get(url)
    data = r.json()
    data = data['data']
    SONG_TUPLE = []

    if data:
        for i in range(0, len(data)):
            song_obj = DeezerSongs(data[i])
            SONG_TUPLE.append(song_obj)
    return SONG_TUPLE


if __name__ == '__main__':
    q = input("Enter the query: ")
    dat = searchSong(q)
    for i in range(0, len(dat)):
        print(dat[i].collection_name)



"""
Handle all the util functions for working with YouTube Music.
"""

from ytmusicapi import YTMusic
from simber import Logger

from ytmdl.exceptions import ExtractError


# Create logger
logger = Logger("ytmusic")


def get_title_from_ytmusic(videoId: str) -> str:
    """
    Get the title of the song from the videoID.

    Youtube actually supports Youtube videoId's
    that can be found in YouTube Music with the
    proper song title.

    This is way more effective and correct than
    extracting the title from the YouTube video.
    """
    ytmusic = YTMusic()

    details = ytmusic.get_song(videoId=videoId)

    # Check if error occured
    if details["playabilityStatus"]["status"] != "OK":
        raise ExtractError(videoId)

    try:
        return details["videoDetails"]["title"]
    except KeyError:
        raise ExtractError(videoId)



#!/usr/bin/env python3
"""Module for caching songs.

    Nishan Pantha(c) 2017-18
    NISH1001@github.com
"""

from glob import glob
import os
from ytmdl.stringutils import (
    remove_multiple_spaces, remove_punct, compute_jaccard, remove_stopwords,
    check_keywords
)
from ytmdl.defaults import DEFAULT
from ytmdl.prepend import PREPEND
from simber import Logger
from colorama import Fore, Style

logger = Logger("cache")


class Cache:
    """The main caching component."""

    def __init__(self, directory=None):
        """Init the stuff."""
        if directory is None:
            directory = DEFAULT.SONG_DIR
            # Check the dir only if special characters are present
            if '$' in directory:
                directory = directory.split("$")[0]
        directory = os.path.expanduser(directory)
        self.directory = directory
        self.max_depth = 4

    def _get_files(self, song_name):
        """
        List all the files in the passed dir.
        """
        files = []
        for depth in range(0, self.max_depth):
            pattern = '{}{}{}*'.format(
                self.directory,
                '{}'.format(os.sep).join(depth * ['*']),
                song_name
            )
            files.extend(glob(pattern))

        return files

    def search(self, song_name):
        logger.info(
            "Searching to see if already present in {}".format(
                self.directory)
        )
        return len(self._get_files(song_name))


def main(SONG_NAME=''):
    """Run on program call."""
    cache = Cache()
    match = cache.search(SONG_NAME)
    if match:
        PREPEND(1)
        print(Fore.MAGENTA, end='')
        print('{} '.format(SONG_NAME), end='')
        print(Style.RESET_ALL, end='')
        print('found.')
        while True:
            choice = input('Do you still want to continue[y/n]')
            choice = choice.lower()
            if choice == 'y' or choice == 'Y':
                return True
            elif choice == 'n' or choice == 'N':
                return False
    else:
        return True


if __name__ == "__main__":
    main("Cradles")



"""Module to handle meta fetch from musicbrainz"""

import musicbrainzngs
from ytmdl.__version__ import __version__

musicbrainzngs.set_useragent(
    "ytmdl",
    __version__,
    "https://github.com/deepjyoti30/ytmdl"
)


class MusicBrainzSong():
    """Class to store gaana song tags."""

    def __init__(self, SONG):
        """SONG is supposed to be a dict."""
        self.track_name = SONG['title']
        self.release_date = self.__get_date(SONG)
        self.artist_name = SONG['artist-credit'][0]['name']
        self.provider = "musicbrainz"
        self.collection_name = SONG['release-list'][0]['title']
        self.primary_genre_name = ""  # Seems musicbrainz doesn't provide genre
        self.track_number = SONG['release-list'][0]['medium-list'][0]['track-list'][0]['number']
        self.artwork_url_100 = ""
        self.track_time = self.__get_length(SONG)

        # Below will be used to fetch extra data
        self.__release_id = SONG['release-list'][0]['id']

    def __get_length(self, song):
        """Try to extract the length of the song"""
        try:
            return self._convert_time(song['length'])
        except KeyError:
            return ""

    def __get_date(self, song):
        """Try to extract the length of the song"""
        try:
            return song['release-list'][0]['date']
        except KeyError:
            return ""

    def _convert_time(self, duration):
        in_min = int(duration)
        in_time = int(in_min / 60) + (0.01 * (in_min % 60))
        return in_time

    @property
    def release_id(self) -> str:
        return self.__release_id


def search_song(query, lim=25):
    """Search the song using the API and return the
    results accordingly
    """
    recordings = musicbrainzngs.search_recordings(query)

    if not recordings:
        return []

    data = []

    recordings_results = recordings["recording-list"]
    for recording in recordings_results:
        data.append(MusicBrainzSong(recording))

    return data


def get_more_data(song):
    """Get extra data for the song like cover art"""
    id = song.release_id

    cover_art = musicbrainzngs.get_image_list(id)
    song.artwork_url_100 = cover_art["images"][0]["image"]

    return song


if __name__ == "__main__":
    data = search_song("Cradles")
    for i in data:
        print("{}:{}".format(i.track_name, i.artist_name))



"""Functions used in setting up the config file are defined here."""

import os
import re
from simber import Logger
from xdg.BaseDirectory import xdg_config_home


config_text = '''#*****************************************#
#*-------------config for ytmdl-----------#
#
#-----------------------------------------#
#------PLEASE DON\'T LEAVE ANY BLANK LINES---#
#-----------------------------------------#
#
# To change defaults just remove the hash(#)
# from the beginning of the line.
#
#*****************************************
# The SONG_DIR is the directory where all the songs will be saved.
# In order to change it, simply remove the hash from beginning
# And change the path to your desired one.
# In case the path has spaces in in, include it in a " "
# Following is a simple folder path example
#
#SONG_DIR = "path/to/your/desired/folder"
#
#************--------ADVANCED-------*********
# If you want to save the song in custom folders than those can be
# added to the name like the following example.
# The possible values are following
#
# Artist --> Song Artist
# Album  --> Song Album Name
# Title  --> Song Name
# Genre  --> Song Genre
# TrackNumber --> Song Number in the album
# ReleaseDate --> Song Release date
#
# Following is an example of the format
#SONG_DIR = "/home/user/Music$Artist->Album->Title"
#
#*****************************************#
# The QUALITY is the quality of the song in kbps
# By default it is set to 320kbps
# In case you want to change it to something else,
# Uncomment the following line and change it
#
# Supported values are 320 and 192
#
#QUALITY = "320"
#
#*****************************************#
# The METADATA_PROVIDERS value is a comma separated
# values that specifies wich API providers to use for getting
# the song metadata. Available values right now are:
#
# ---------
# "{{supported_providers}}".
# ---------
#
# Please check the github page of ytmdl for more information.
#
#METADATA_PROVIDERS = "itunes, spotify, gaana"
#
#*****************************************#
# The DEFAULT_FORMAT denotes what to use as default for downloading.
# Available values are:
#
# ---------
# "{{supported_formats}}"
# ---------
#
#DEFAULT_FORMAT = "mp3"
#
#*****************************************#
# What to do when something goes wrong while adding the metadata. This can
# happen for various reasons like lack of metadata or network issues.
# Available options are:
#
# ---------
# "{{supported_on_error_options}}"
# ---------
#
#ON_META_ERROR = "exit"
#****************************************#
# ITUNES SETTINGS
# Which region to use while searching from Itunes.
#
#
# Country where the song is to be searched for in Itunes.
# The country helps if you're looking for regional songs in
# which case Itunes orders the results accordingly.
#
# By default, 'US' will be used.
#
#ITUNES_COUNTRY = "US"
#
# Country where the song is to be searched for in Spotify.
# The country helps if you're looking for regional songs in
# which case Spotify orders the results accordingly.
#
# By default, 'US' is used
#
#SPOTIFY_COUNTRY = "US"
#'''


logger = Logger("config")


class DEFAULTS:
    """Some default stuff defined."""

    def __init__(self):
        # The home dir
        self.HOME_DIR = os.path.expanduser('~')

        # The default song dir
        self.SONG_DIR = self._get_music_dir()

        # The temp dir
        self.SONG_TEMP_DIR = os.path.join(self.SONG_DIR, 'ytmdl')

        # The default song quality
        self.SONG_QUALITY = '320'

        # The config path
        self.CONFIG_PATH = os.path.join(xdg_config_home, 'ytmdl')

        # The default metadata providers
        self.METADATA_PROVIDERS = ['itunes', 'spotify', 'gaana']

        # The available metadata providers
        self.AVAILABLE_METADATA_PROVIDERS = self.METADATA_PROVIDERS + \
            ['deezer', 'lastfm', 'musicbrainz', 'saavn']  # add new ones here

        self.VALID_FORMATS = ['mp3', 'm4a', 'opus']

        self.DEFAULT_FORMAT = 'mp3'

        self.ON_ERROR_OPTIONS = ['exit', 'skip', 'manual']

        self.ON_ERROR_DEFAULT = 'exit'

        # Itunes related settings
        self.ITUNES_COUNTRY_DEFAULT = "US"

        # Spotify country
        self.SPOTIFY_COUNTRY_DEFAULT = "US"

    def _get_music_dir(self):
        """Get the dir the file will be saved to."""
        # The first preference will be ~/Music.
        # If that is not present, try checking the XDG_MUSIC_DIR
        # If still not, then use the current directory.
        music_dir = self._get_xdg_dir()

        if music_dir is None:
            music_dir = os.path.join(self.HOME_DIR, 'Music')

        if not os.path.exists(music_dir):
            music_dir = os.getcwd()

        return music_dir

    def _get_xdg_dir(self):
        """Get the xdg dir."""
        file_path = os.path.expanduser('~/.config/user-dirs.dirs')

        if not os.path.exists(file_path):
            return None

        with open(file_path, 'r') as RSTREAM:
            data = RSTREAM.read()
            path = re.findall(r'\nXDG_MUSIC_DIR.*?\n', str(data))
            if not path:
                return None
            path = re.sub(r'\n|XDG_MUSIC_DIR|=|"', '', path[0])
            path = os.path.expandvars(path)
            return path


def render_config_template() -> str:
    """Render the config template in order ot get the updated
    config
    """
    defaults_obj = DEFAULTS()
    providers = ", ".join(defaults_obj.AVAILABLE_METADATA_PROVIDERS)
    formats = ", ".join(defaults_obj.VALID_FORMATS)
    on_error_options = ", ".join(defaults_obj.ON_ERROR_OPTIONS)

    KEYWORD_MAP = {
        '"{{supported_providers}}"': providers,
        '"{{supported_formats}}"': formats,
        '"{{supported_on_error_options}}"': on_error_options
    }

    rendered_content = config_text

    for keyword, value in KEYWORD_MAP.items():
        rendered_content = rendered_content.replace(keyword, value)

    return rendered_content


def make_config():
    """Copy the config file to .config folder."""
    # Remove the current config from SONG_TEMP_DIR
    config_path = os.path.join(DEFAULTS().CONFIG_PATH, 'config')

    # Check if the ytmdl folder is present in config
    if not os.path.isdir(DEFAULTS().CONFIG_PATH):
        # Make the ytmdl folder
        os.makedirs(DEFAULTS().CONFIG_PATH, exist_ok=True)

    elif os.path.isfile(config_path):
        os.remove(config_path)

    # Check if the ytmdl folder is present in Music directory
    if not os.path.isdir(DEFAULTS().SONG_TEMP_DIR):
        # Make the ytmdl folder
        os.makedirs(DEFAULTS().SONG_TEMP_DIR)

    # Now write the config text to config file
    with open(config_path, 'w') as write_config:
        write_config.write(render_config_template())


def checkConfig():
    """Need to check the config to see if defaults are changed.

    The config will be saved in the .config folder.
    """
    # Try to see if the config is present in the SONG_TEMP_DIR

    if os.path.isdir(DEFAULTS().CONFIG_PATH):
        DIR_CONTENTS = os.listdir(DEFAULTS().CONFIG_PATH)
    else:
        return False

    if 'config' not in DIR_CONTENTS:
        make_config()
        return True
    else:
        return True


def check_config_setup():
    """
    Method to check if the config file is setup.

    This is different from the above method because
    it will check if the config is setup and not do
    anything about it.

    The return value can be used to perform on config
    setup etc.
    """
    # There are two possibilities that might indicate
    # that the config is not setup.
    DEFAULT_CONF_PATH = DEFAULTS().CONFIG_PATH

    # Check if ytmdl config directory is present
    if not os.path.isdir(DEFAULT_CONF_PATH):
        return False

    # Check if config file is present
    if not os.path.isfile(os.path.join(DEFAULT_CONF_PATH, 'config')):
        return False

    # Else it's probably setup
    return True


def checkValidity(keyword, value):
    """Check if the user specified value in config is possible."""
    if keyword == 'SONG_DIR':
        # In this case check if $ and -> are present
        # If they are then only check if the base dir exists
        if '$' in value:
            pos = value.find('$')
            value = value[:pos]
        return os.path.isdir(os.path.expanduser(value))
    elif keyword == 'QUALITY':
        # Possible values that QUALITY can take
        possQ = ['320', '192']
        return value in possQ
    elif keyword == 'DEFAULT_FORMAT':
        possF = DEFAULTS().VALID_FORMATS
        return value in possF
    elif keyword == 'METADATA_PROVIDERS':
        # Possible values that METADATA_PROVIDERS can take
        possM = DEFAULTS().AVAILABLE_METADATA_PROVIDERS
        if not value:
            logger.warning(
                "Metadata provider value is empty. \
                    Default values will be used.")
            return False
        new_val = value.replace(' ', '').split(',')

        # Even if one is valid, return true
        for provider in new_val:
            if provider in possM:
                return True
        return False
    elif keyword == "ON_META_ERROR":
        if not value:
            logger.warning("On meta error value is empty. \
                    Default will be used")
            return False

        if value not in DEFAULTS().ON_ERROR_OPTIONS:
            return False
        return True

    elif keyword == "ITUNES_COUNTRY":
        if not value:
            logger.warning("Itunes Country value is empty. \
                Default will be used")
            return False

        # TODO: Perhaps check if valid country
        return True

    elif keyword == "SPOTIFY_COUNTRY":
        if not value:
            logger.warning("Spotify Country value is empty. \
                Default will be used")
            return False

        # TODO: Perhaps check if valid country
        return True


def retDefault(keyword):
    """Return the DEFAULT value of keyword."""
    if keyword == 'QUALITY':
        return DEFAULTS().SONG_QUALITY
    elif keyword == 'DEFAULT_FORMAT':
        return DEFAULTS().DEFAULT_FORMAT
    elif keyword == 'SONG_DIR':
        return DEFAULTS().SONG_DIR
    elif keyword == 'METADATA_PROVIDERS':
        return DEFAULTS().METADATA_PROVIDERS
    elif keyword == "ON_META_ERROR":
        return DEFAULTS().ON_ERROR_DEFAULT
    elif keyword == "ITUNES_COUNTRY":
        return DEFAULTS().ITUNES_COUNTRY_DEFAULT
    elif keyword == "SPOTIFY_COUNTRY":
        return DEFAULTS().SPOTIFY_COUNTRY_DEFAULT


def GIVE_DEFAULT(self, keyword):
    """Check if the user has uncommented the config and added something.

    If possible get what is changed, else return the default value.
    """
    # Check If the config is already present in SONG_TEMP_DIR
    if not checkConfig():
        return retDefault(keyword)
    else:
        # Then read from it
        READ_STREAM = open(os.path.join(DEFAULTS().CONFIG_PATH, 'config'), 'r')

        while True:
            line = READ_STREAM.readline()
            if not line:
                return retDefault(keyword)
            if line[0] != '#' and keyword in line:
                # Get the position of =
                index_equal = line.index('=')
                if line[index_equal + 1] == ' ':
                    newDEFAULT = line[index_equal + 2:]
                else:
                    newDEFAULT = line[index_equal + 1:]

                # Remove the "
                newDEFAULT = newDEFAULT.replace('"', '')
                newDEFAULT = newDEFAULT.replace("'", '')
                # Check if the line has a \n in it
                if "\n" in line:
                    newDEFAULT = newDEFAULT.replace('\n', '')

                if checkValidity(keyword, newDEFAULT):
                    return newDEFAULT
                else:
                    if newDEFAULT:
                        logger.warning(
                            "{}: is invalid for option {}.".format(newDEFAULT, keyword))
                    return retDefault(keyword)


if __name__ == '__main__':
    # Create the config in the examples directory
    try:
        with open("examples/config", 'w') as write_config:
            write_config.write(render_config_template())
    except Exception:
        print("You need to run this script in root directory of ytmdl")



#!/usr/bin/env python3
"""Setup ytmdl."""

import setuptools
from os import path, name
from warnings import warn

with open("README.md", "r") as fh:
    long_description = fh.read()

exec(open("ytmdl/__version__.py").read())

req_pkgs = [
    'youtube_dl',
    'mutagen',
    'itunespy',
    'requests',
    'colorama',
    'bs4',
    'downloader-cli',
    'pyxdg',
    'ffmpeg-python',
    'pysocks',
    'unidecode',
    'youtube-search-python',
    'pyDes',
    'urllib3',
    'simber',
    'rich',
    'musicbrainzngs',
    'ytmusicapi',
    'spotipy'
]


extra_features = {
    'noise-clean': ['inaSpeechSegmenter', 'tensorflow']
}


# Add the distributable files
file_map = [
    ('/etc/bash_completion.d', 'ytmdl.bash'),
    ('/usr/share/zsh/functions/Completion/Unix', 'ytmdl.zsh')
]

data_files = []
for dirname, filename in file_map:
    if not path.exists(filename):
        warn("%s does not exist, skipping." % filename)
    else:
        data_files.append((dirname, [filename]))

params = {
    'data_files': data_files,
}

params["scripts"] = ['bin/ytmdl']

# Install a bat script for Windows users
if name == 'nt':
    params["scripts"].append("bin/ytmdl.bat")


if __name__ == '__main__':
    setuptools.setup(
        name="ytmdl",
        version=__version__,
        author="Deepjyoti Barman",
        author_email="deep.barma30@gmail.com",
        description="Youtube Music Downloader",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/deepjyoti30/ytmdl",
        packages=setuptools.find_packages(),
        classifiers=(
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Topic :: Multimedia :: Sound/Audio",
        ),
        python_requires=">=3.*",
        install_requires=req_pkgs,
        setup_requires=req_pkgs,
        extras_require=extra_features,
        **params
    )



"""Handle data extraction from the Last.FM API

This module will handle everything related to extracting
the required data from the Last.FM source.
"""

from requests import get

API_BASE = "http://ws.audioscrobbler.com/2.0/"
API_KEY = "865e60e7cf58e028c063cb1230c95e5e"


class LastFMSongs():
    """Class to store Last FM song tags."""

    def __init__(self, SONG):
        """SONG is supposed to be a dict.

        All the empty attributes will be filled up in
        the extra feature fetching function.
        """
        self.track_name = SONG['name']
        self.artist_name = SONG['artist']
        self.provider = 'lastfm'
        self.track_number = "1"
        self.collection_name = ""
        self.release_date = ""
        self.artwork_url_100 = SONG["image"][-1]["#text"]
        self.track_time = ""
        self.primary_genre_name = "N/A"

    def _convert_time(self, duration):
        """duration is in ms"""
        in_sec = int(int(duration) / 1000)
        in_time = int(in_sec / 60) + (0.01 * (in_sec % 60))
        return in_time


def get_more_data(song):
    """song is a LastFMSong object.

    We can easily fetch the artist and track name and
    according get the extra required data from the API.
    """
    headers = {
        "User-Agent": "ytmdl"
    }
    payload = {
        "api_key": API_KEY,
        "method": "track.getInfo",
        "track": song.track_name,
        "artist": song.artist_name,
        "format": "json"
    }

    response = get(API_BASE, headers=headers, params=payload)

    # TODO: Add a check to exit if the response code is not 200

    track_details = response.json()

    # Update the songs attributes
    song.track_number = 1

    try:
        song.collection_name = track_details["track"]["album"]["title"]
        song.track_time = song._convert_time(
            track_details["track"]["duration"])
        song.release_date = track_details["track"]["wiki"]["published"]
    except KeyError:
        # This happens because last.fm do not have consistent data for some songs
        # Just ignore this errors if they occur.
        pass

    return song


def searchSong(query, lim=40):
    """Search the song using the last.fm API
    """
    headers = {
        "User-Agent": "ytmdl"
    }
    payload = {
        "api_key": API_KEY,
        "method": "track.search",
        "track": query,
        "format": "json"
    }

    data = []

    response = get(API_BASE, headers=headers, params=payload)

    if response.status_code != 200:
        print(response.status_code)
        return data

    for song in response.json()["results"]["trackmatches"]["track"]:
        data.append(LastFMSongs(song))

    return data


if __name__ == "__main__":
    songs = searchSong("Believer")
    song = get_more_data(songs[0])

    print(song.track_name, song.artist_name)
    print(song.collection_name, song.track_time, song.release_date)
